@@ -22,6 +22,7 @@
 #include <config.h>
 #endif
 
+#include <gdk/gdkkeysyms.h>
 #include <glib/gi18n.h>
 #include <glade/glade.h>
 #include <gnome-keyring.h>
@@ -30,6 +31,7 @@
 #include "vinagre-notebook.h"
 #include "vinagre-tab.h"
 #include "vinagre-utils.h"
+#include "vinagre-prefs.h"
 
 #define VINAGRE_TAB_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), VINAGRE_TYPE_TAB, VinagreTabPrivate))
 
@@ -45,6 +47,7 @@ struct _VinagreTabPrivate
   gboolean           save_password;
   guint32            keyring_item_id;
   VinagreTabState    state;
+  gchar             *clipboard_str;
 };
 
 G_DEFINE_TYPE(VinagreTab, vinagre_tab, GTK_TYPE_VBOX)
@@ -109,6 +112,7 @@ vinagre_tab_window_state_cb (GtkWidget           *widget,
       gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (tab->priv->scroll),
 				      GTK_POLICY_AUTOMATIC,
 				      GTK_POLICY_AUTOMATIC);
+      vinagre_notebook_show_hide_tabs (tab->priv->nb);
     }
 
   return FALSE;
@@ -148,6 +152,7 @@ vinagre_tab_set_property (GObject      *object,
     {
       case PROP_CONN:
         tab->priv->conn = g_value_get_object (value);
+	g_object_set_data (G_OBJECT (tab->priv->conn), VINAGRE_TAB_KEY, tab);
         break;
       case PROP_WINDOW:
         tab->priv->window = g_value_get_object (value);
@@ -176,7 +181,10 @@ vinagre_tab_finalize (GObject *object)
   					tab);
   g_object_unref (tab->priv->conn);
   tab->priv->conn = NULL;
-	
+
+  g_free (tab->priv->clipboard_str);
+  tab->priv->clipboard_str = NULL;
+
   G_OBJECT_CLASS (vinagre_tab_parent_class)->finalize (object);
 }
 
@@ -330,22 +338,45 @@ vnc_auth_unsupported_cb (VncDisplay *vnc, guint auth_type, VinagreTab *tab)
   vinagre_notebook_remove_tab (tab->priv->nb, tab);
 }
 
+/* text was actually requested */
+static void
+copy_cb (GtkClipboard     *clipboard,
+         GtkSelectionData *data,
+	 guint             info,
+	 VinagreTab       *tab)
+{
+  gtk_selection_data_set_text (data, tab->priv->clipboard_str, -1);
+}
+
 static void
 vnc_server_cut_text_cb (VncDisplay *vnc, const gchar *text, VinagreTab *tab)
 {
   GtkClipboard *cb;
-  gchar *out;
   gsize a, b;
+  GtkTargetEntry targets[] = {
+				{"UTF8_STRING", 0, 0},
+				{"COMPOUND_TEXT", 0, 0},
+				{"TEXT", 0, 0},
+				{"STRING", 0, 0},
+			     };
 
   if (!text)
     return;
 
-  out = g_convert (text, -1, "utf-8", "iso8859-1", &a, &b, NULL);
-  if (out) {
-    cb = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
-    gtk_clipboard_set_text (cb, out, -1);
-    g_free (out);
-  }
+  g_free (tab->priv->clipboard_str);
+  tab->priv->clipboard_str = g_convert (text, -1, "utf-8", "iso8859-1", &a, &b, NULL);
+
+  if (tab->priv->clipboard_str)
+    {
+      cb = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
+
+      gtk_clipboard_set_with_owner (cb,
+				    targets,
+				    G_N_ELEMENTS(targets),
+				    (GtkClipboardGetFunc) copy_cb,
+				    NULL,
+				    G_OBJECT (tab));
+    }
 }
 
 static void
@@ -519,21 +550,36 @@ static void
 vnc_authentication_cb (VncDisplay *vnc, GValueArray *credList, VinagreTab *tab)
 {
   gchar *password;
+  int i;
 
-  password = vinagre_tab_find_password (tab);
-  if (!password)
-    {
-      password = ask_password (tab);
-      if (!password) {
-        vinagre_notebook_remove_tab (tab->priv->nb, tab);
-        return;
+  if (credList == NULL)
+    return;
+
+  for (i = 0; i < credList->n_values; i++) {
+    switch (g_value_get_enum (&credList->values[i]))
+      {
+        case VNC_DISPLAY_CREDENTIAL_PASSWORD:
+          password = vinagre_tab_find_password (tab);
+          if (!password)
+            {
+              password = ask_password (tab);
+              if (!password) {
+                vinagre_notebook_remove_tab (tab->priv->nb, tab);
+                return;
+              }
+            }
+
+          vinagre_connection_set_password (tab->priv->conn, password);
+          vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_PASSWORD, password);
+
+          g_free (password);
+          break;
+
+        case VNC_DISPLAY_CREDENTIAL_CLIENTNAME:
+          vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_CLIENTNAME, "vinagre");
+          break;
       }
-    }
-
-  vinagre_connection_set_password (tab->priv->conn, password);
-  vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_PASSWORD, password);
-
-  g_free (password);
+  }
 }
 
 static void vnc_grab_cb (VncDisplay *vnc, VinagreTab *tab)
@@ -575,15 +621,25 @@ static void vnc_ungrab_cb (VncDisplay *vnc, VinagreTab *tab)
     gtk_notebook_set_show_tabs (GTK_NOTEBOOK (tab->priv->nb), TRUE);
 }
 
+static gboolean
+vnc_key_event ( GtkWidget   *widget G_GNUC_UNUSED,
+		GdkEventKey *key G_GNUC_UNUSED,
+		gpointer     data G_GNUC_UNUSED)
+{
+  return TRUE;
+}
+
 static void
 vinagre_tab_init (VinagreTab *tab)
 {
   GtkWidget *viewport;
+  gboolean shared;
 
   tab->priv = VINAGRE_TAB_GET_PRIVATE (tab);
   tab->priv->save_password = FALSE;
   tab->priv->keyring_item_id = 0;
   tab->priv->state = VINAGRE_TAB_STATE_INITIALIZING;
+  tab->priv->clipboard_str = NULL;
 
   /* Create the scrolled window */
   tab->priv->scroll = gtk_scrolled_window_new (NULL, NULL);
@@ -603,6 +659,12 @@ vinagre_tab_init (VinagreTab *tab)
   viewport = gtk_bin_get_child (GTK_BIN (tab->priv->scroll));
   gtk_viewport_set_shadow_type(GTK_VIEWPORT (viewport), GTK_SHADOW_NONE);
 
+  g_object_get (vinagre_prefs_get_default (),
+		"shared-flag", &shared,
+		NULL);
+  vnc_display_set_shared_flag (VNC_DISPLAY (tab->priv->vnc),
+			       shared);
+
   g_signal_connect (tab->priv->vnc,
 		    "vnc-connected",
 		    G_CALLBACK (vnc_connected_cb),
@@ -653,6 +715,16 @@ vinagre_tab_init (VinagreTab *tab)
 		    G_CALLBACK (vnc_bell_cb),
 		    tab);
 
+  g_signal_connect (tab->priv->vnc,
+		    "key-press-event",
+		    G_CALLBACK (vnc_key_event),
+		    NULL);
+
+  g_signal_connect (tab->priv->vnc,
+		    "key-release-event",
+		    G_CALLBACK (vnc_key_event),
+		    NULL);
+
   gtk_widget_show_all (GTK_WIDGET (tab));
 }
 
@@ -781,6 +853,16 @@ vinagre_tab_take_screenshot (VinagreTab *tab)
   g_free (name);
 }
 
+void
+vinagre_tab_send_ctrlaltdel (VinagreTab *tab)
+{
+  guint keys[] = { GDK_Control_L, GDK_Alt_L, GDK_Delete };
+
+  g_return_if_fail (VINAGRE_IS_TAB (tab));
+
+  vnc_display_send_keys_ex (VNC_DISPLAY (tab->priv->vnc), keys, sizeof (keys) / sizeof (keys[0]), VNC_DISPLAY_KEY_EVENT_CLICK);
+}
+
 void
 vinagre_tab_paste_text (VinagreTab *tab, const gchar *text)
 {
@@ -836,4 +918,16 @@ vinagre_tab_get_state (VinagreTab *tab)
   return tab->priv->state;
 }
 
+VinagreTab *
+vinagre_tab_get_from_connection (VinagreConnection *conn)
+{
+  gpointer res;
+	
+  g_return_val_if_fail (VINAGRE_IS_CONNECTION (conn), NULL);
+	
+  res = g_object_get_data (G_OBJECT (conn), VINAGRE_TAB_KEY);
+	
+  return (res != NULL) ? VINAGRE_TAB (res) : NULL;
+}
+
 /* vim: ts=8 */
