@@ -32,7 +32,7 @@
 #include "vinagre-window.h"
 #include "vinagre-notebook.h"
 #include "vinagre-fav.h"
-#include "vinagre-prefs-manager.h"
+#include "vinagre-prefs.h"
 #include "vinagre-utils.h"
 #include "vinagre-bookmarks.h"
 #include "vinagre-ui.h"
@@ -47,7 +47,7 @@
 G_DEFINE_TYPE(VinagreWindow, vinagre_window, GTK_TYPE_WINDOW)
 
 static void
-vinagre_window_finalize (GObject *object)
+vinagre_window_dispose (GObject *object)
 {
   VinagreWindow *window = VINAGRE_WINDOW (object);
 
@@ -63,6 +63,37 @@ vinagre_window_finalize (GObject *object)
       window->priv->manager = NULL;
     }
 
+  if (window->priv->signal_notebook != 0)
+    {
+      g_signal_handler_disconnect (window->priv->notebook,
+				   window->priv->signal_notebook);
+      window->priv->signal_notebook = 0;
+    }
+
+  if (window->priv->signal_clipboard != 0)
+    {
+      GtkClipboard  *cb = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);;
+
+      g_signal_handler_disconnect (cb,
+				   window->priv->signal_clipboard);
+      window->priv->signal_clipboard = 0;
+    }
+
+  if (window->priv->update_recents_menu_ui_id != 0)
+    {
+      GtkRecentManager *recent_manager = gtk_recent_manager_get_default ();
+
+      g_signal_handler_disconnect (recent_manager,
+				   window->priv->update_recents_menu_ui_id);
+      window->priv->update_recents_menu_ui_id = 0;
+    }
+
+  G_OBJECT_CLASS (vinagre_window_parent_class)->dispose (object);
+}
+
+static void
+vinagre_window_finalize (GObject *object)
+{
   G_OBJECT_CLASS (vinagre_window_parent_class)->finalize (object);
 }
 
@@ -71,20 +102,8 @@ vinagre_window_delete_event (GtkWidget   *widget,
 			     GdkEventAny *event)
 {
   VinagreWindow *window = VINAGRE_WINDOW (widget);
-  GtkClipboard  *cb;
-
-  cb = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
-
-  if (window->priv->signal_notebook > 0)
-    g_signal_handler_disconnect (window->priv->notebook,
-				 window->priv->signal_notebook);
-
-  if (window->priv->signal_clipboard > 0)
-    g_signal_handler_disconnect (cb,
-				 window->priv->signal_clipboard);
 
   vinagre_window_close_all_tabs (window);
-  gtk_main_quit ();
 
   if (GTK_WIDGET_CLASS (vinagre_window_parent_class)->delete_event)
     return GTK_WIDGET_CLASS (vinagre_window_parent_class)->delete_event (widget, event);
@@ -124,7 +143,7 @@ vinagre_window_show_hide_controls (VinagreWindow *window)
       if (window->priv->statusbar_visible)
         gtk_widget_show_all (window->priv->statusbar);
 
-      gtk_notebook_set_show_tabs (GTK_NOTEBOOK (window->priv->notebook), TRUE);
+      vinagre_notebook_show_hide_tabs (VINAGRE_NOTEBOOK (window->priv->notebook));
       gtk_notebook_set_show_border (GTK_NOTEBOOK (window->priv->notebook), TRUE);
     }
 }
@@ -133,21 +152,21 @@ static gboolean
 vinagre_window_state_event_cb (GtkWidget *widget,
 			       GdkEventWindowState *event)
 {
-  VinagreWindow *window;
+  VinagreWindow *window = VINAGRE_WINDOW (widget);
+
+  window->priv->window_state = event->new_window_state;
+  g_object_set (vinagre_prefs_get_default (),
+		"window-state", window->priv->window_state,
+		NULL);
 
   if ((event->changed_mask & GDK_WINDOW_STATE_FULLSCREEN) == 0)
     return FALSE;
 
-  window = VINAGRE_WINDOW (widget);
-
   if (event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN)
     window->priv->fullscreen = TRUE;
   else
     window->priv->fullscreen = FALSE;
 
-  window->priv->window_state = event->new_window_state;
-  vinagre_prefs_manager_set_window_state (window->priv->window_state);
-
   vinagre_window_show_hide_controls (window);
 
   if (GTK_WIDGET_CLASS (vinagre_window_parent_class)->window_state_event)
@@ -165,8 +184,10 @@ vinagre_window_configure_event (GtkWidget         *widget,
   window->priv->width  = event->width;
   window->priv->height = event->height;
 
-  vinagre_prefs_manager_set_window_size (window->priv->width,
-					 window->priv->height);
+  g_object_set (vinagre_prefs_get_default (),
+		"window-width", window->priv->width,
+		"window-height", window->priv->height,
+		NULL);
 
   return GTK_WIDGET_CLASS (vinagre_window_parent_class)->configure_event (widget, event);
 }
@@ -196,7 +217,8 @@ vinagre_window_class_init (VinagreWindowClass *klass)
   GObjectClass *object_class    = G_OBJECT_CLASS (klass);
   GtkWidgetClass *widget_class  = GTK_WIDGET_CLASS (klass);
 
-  object_class->finalize     = vinagre_window_finalize;
+  object_class->finalize = vinagre_window_finalize;
+  object_class->dispose  = vinagre_window_dispose;
 
   widget_class->key_press_event    = vinagre_window_key_press_cb;
   widget_class->window_state_event = vinagre_window_state_event_cb;
@@ -289,7 +311,8 @@ activate_recent_cb (GtkRecentChooser *action, VinagreWindow *window)
     g_free (error);
 }
 
-static void update_recent_connections (VinagreWindow *window)
+static void
+update_recent_connections (VinagreWindow *window)
 {
   VinagreWindowPrivate *p = window->priv;
 
@@ -396,6 +419,8 @@ create_menu_bar_and_toolbar (VinagreWindow *window,
   g_object_set (action, "is_important", TRUE, NULL);
   action = gtk_action_group_get_action (action_group, "MachineTakeScreenshot");
   g_object_set (action, "is_important", TRUE, NULL);
+  action = gtk_action_group_get_action (action_group, "MachineSendCtrlAltDel");
+  g_object_set (action, "is_important", TRUE, NULL);
 
   /* now load the UI definition */
   gtk_ui_manager_add_ui_from_file (manager, vinagre_utils_get_ui_xml_filename (), &error);
@@ -461,11 +486,11 @@ create_menu_bar_and_toolbar (VinagreWindow *window,
   update_recent_connections (window);
 
   recent_manager = gtk_recent_manager_get_default ();
-  g_signal_connect (recent_manager,
+  window->priv->update_recents_menu_ui_id = g_signal_connect (
+		    recent_manager,
 		    "changed",
 		    G_CALLBACK (recent_manager_changed),
 		    window);
-
 }
 
 void
@@ -490,7 +515,9 @@ fav_panel_size_allocate (GtkWidget     *widget,
 {
   window->priv->side_panel_size = allocation->width;
   if (window->priv->side_panel_size > 0)
-    vinagre_prefs_manager_set_side_panel_size (window->priv->side_panel_size);
+    g_object_set (vinagre_prefs_get_default (),
+		  "side-panel-size", window->priv->side_panel_size,
+		  NULL);
 }
 
 static void
@@ -681,7 +708,9 @@ create_side_panel (VinagreWindow *window)
 		   FALSE, 
 		   FALSE);
 
-  window->priv->side_panel_size = vinagre_prefs_manager_get_side_panel_size ();
+  g_object_get (vinagre_prefs_get_default (),
+		"side-panel-size", &(window->priv->side_panel_size),
+		NULL);
   gtk_paned_set_position (GTK_PANED (window->priv->hpaned), window->priv->side_panel_size);
 
   g_signal_connect (window->priv->fav_panel,
@@ -705,6 +734,7 @@ init_widgets_visibility (VinagreWindow *window)
   gint w, h;
   GtkAction *action;
   gboolean visible;
+  VinagrePrefs *prefs = vinagre_prefs_get_default ();
 
   /* Remove and Edit bookmarks starts disabled */
   action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
@@ -715,64 +745,32 @@ init_widgets_visibility (VinagreWindow *window)
 					"BookmarksEdit");
   gtk_action_set_sensitive (action, FALSE);
 
-  /* fav panel visibility */
+  /* side panel visibility */
   action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
 					"ViewSidePanel");
-  visible = vinagre_prefs_manager_get_side_pane_visible ();
+  g_object_get (prefs, "side-panel-visible", &visible, NULL);
   if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)) != visible)
     gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), visible);
 
   /* toolbar visibility */
   action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
 					"ViewToolbar");
-  visible = vinagre_prefs_manager_get_toolbar_visible ();
+  g_object_get (prefs, "toolbar-visible", &visible, NULL);
   if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)) != visible)
     gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), visible);
 
   /* statusbar visibility */
   action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
 					"ViewStatusbar");
-  visible = vinagre_prefs_manager_get_side_pane_visible ();
+  g_object_get (prefs, "statusbar-visible", &visible, NULL);
   if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)) != visible)
     gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), visible);
 
-  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
-				vinagre_prefs_manager_get_statusbar_visible ());
-
-//  if (vinagre_prefs_manager_get_side_pane_visible ())
-//    gtk_widget_show_all (window->priv->fav_panel);
-//  else
-//    {
-//      gtk_widget_hide (window->priv->fav_panel);
-//printf ("hiding... was visible: %d\n", GTK_WIDGET_VISIBLE(window->priv->fav_panel));
-
-//      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
-//    }
-
-  /* toolbar visibility */
-//  action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
-//					"ViewToolbar");
-//  if (vinagre_prefs_manager_get_toolbar_visible ())
-//    gtk_widget_show_all (window->priv->toolbar);
-//  else
-//    {
-//      gtk_widget_hide (window->priv->toolbar);
-//      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
-//   }
-
-  /* toolbar visibility */
-//  action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
-//					"ViewStatusbar");
-//  if (vinagre_prefs_manager_get_statusbar_visible ())
-//    gtk_widget_show_all (window->priv->statusbar);
-//  else
-//    {
- //     gtk_widget_hide (window->priv->statusbar);
-//      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
- //   }
-
-  state = vinagre_prefs_manager_get_window_state ();
-  vinagre_prefs_manager_get_window_size (&w, &h);
+  g_object_get (prefs,
+		"window-state", &state,
+		"window-width", &w,
+		"window-height", &h,
+		NULL);
 
   gtk_window_set_default_size (GTK_WINDOW (window), w, h);
 
@@ -941,13 +939,18 @@ vinagre_window_clipboard_cb (GtkClipboard *cb, GdkEvent *event, VinagreWindow *w
 {
   gchar *text;
 
-  text = gtk_clipboard_wait_for_text (cb);
-  if (text) {
-    if (window->priv->active_tab)
-      vinagre_tab_paste_text (VINAGRE_TAB (window->priv->active_tab), text);
-    g_free (text);
-  }
+  if (!window->priv->active_tab)
+    return;
 
+  if (gtk_clipboard_get_owner (cb) == G_OBJECT (window->priv->active_tab))
+    return;
+
+  text = gtk_clipboard_wait_for_text (cb);
+  if (!text)
+    return;
+
+  vinagre_tab_paste_text (VINAGRE_TAB (window->priv->active_tab), text);
+  g_free (text);
 }
 
 static void
@@ -1161,4 +1164,60 @@ vinagre_window_get_ui_manager (VinagreWindow *window)
   return window->priv->manager;
 }
 
+static void
+add_connection (VinagreTab *tab, GList **res)
+{
+  VinagreConnection *conn;
+	
+  conn = vinagre_tab_get_conn (tab);
+	
+  *res = g_list_prepend (*res, conn);
+}
+
+/* Returns a newly allocated list with all the connections in the window */
+GList *
+vinagre_window_get_connections (VinagreWindow *window)
+{
+  GList *res = NULL;
+
+  g_return_val_if_fail (VINAGRE_IS_WINDOW (window), NULL);
+	
+  gtk_container_foreach (GTK_CONTAINER (window->priv->notebook),
+			 (GtkCallback)add_connection,
+			  &res);
+			       
+  res = g_list_reverse (res);
+	
+  return res;
+}
+
+VinagreTab *
+vinagre_window_conn_exists (VinagreWindow *window, VinagreConnection *conn)
+{
+  VinagreConnection *c;
+  VinagreTab *tab = NULL;
+  GList *l;
+
+  g_return_val_if_fail (VINAGRE_IS_WINDOW (window), NULL);
+  g_return_val_if_fail (VINAGRE_IS_CONNECTION (conn), NULL);
+
+  l = vinagre_window_get_connections (window);
+
+  while (l != NULL)
+    {
+      c = VINAGRE_CONNECTION (l->data);
+
+      if (!strcmp (vinagre_connection_get_host (conn), vinagre_connection_get_host (c)) &&
+	  vinagre_connection_get_port (conn) == vinagre_connection_get_port (c))
+	{
+	  tab = g_object_get_data (G_OBJECT (c), VINAGRE_TAB_KEY);
+	  break;
+	}
+      l = l->next;
+    }
+  g_list_free (l);
+
+  return tab;
+}
+
 /* vim: ts=8 */
